import os
import json
import time
import yaml
import redis
import asyncio
import datetime
import logging.config

from kafka import KafkaConsumer
from typing import Any
from typing import Dict
from fastapi import Response
from qcloud_cos import CosConfig
from qcloud_cos import CosS3Client

from cfclient.models import *
from cfclient.core import HttpClient
from cfclient.core import TritonClient
from cfclient.utils import get_err_msg
from cfclient.utils import run_algorithm

# This is necessary to register the algorithms
from cfcreator import *


# logging
root = os.path.dirname(__file__)
logging_root = os.path.join(root, "logs", "consumer")
os.makedirs(logging_root, exist_ok=True)
with open(os.path.join(root, "config.yml")) as f:
    now = datetime.datetime.now()
    timestamp = now.strftime("%Y-%m-%d_%H-%M-%S-%f")
    log_path = os.path.join(logging_root, f"{timestamp}.log")
    config = yaml.load(f, Loader=yaml.FullLoader)
    config["handlers"]["file"]["filename"] = log_path
    logging.config.dictConfig(config)

logging.getLogger("aiokafka").disabled = True
logging.getLogger("kafka.conn").disabled = True
logging.getLogger("kafka.client").disabled = True
logging.getLogger("kafka.cluster").disabled = True
logging.getLogger("kafka.coordinator").disabled = True
logging.getLogger("kafka.coordinator.consumer").disabled = True
logging.getLogger("kafka.metrics.metrics").disabled = True
logging.getLogger("kafka.protocol.parser").disabled = True
logging.getLogger("kafka.consumer.fetcher").disabled = True
logging.getLogger("kafka.consumer.group_coordinator").disabled = True
logging.getLogger("kafka.consumer.subscription_state").disabled = True


constants = dict(
    triton_host=None,
    triton_port=8000,
)


# clients
## cos client
config = CosConfig(
    Region=REGION,
    SecretId=SECRET_ID,
    SecretKey=SECRET_KEY,
    Scheme=SCHEME,
)
cos_client = CosS3Client(config)
## http client
http_client = HttpClient()
## triton client
triton_host = constants["triton_host"]
if triton_host is None:
    triton_client = None
else:
    triton_client = TritonClient(url=f"{triton_host}:{constants['triton_port']}")
## collect
clients = dict(
    http=http_client,
    triton=triton_client,
)

redis_client = redis.Redis(**redis_kwargs())
audit_redis_client = redis.Redis(**audit_redis_kwargs())
pending_queue_key = get_pending_queue_key()


# algorithms
loaded_algorithms: Dict[str, IAlgorithm] = {
    k: v(clients) for k, v in algorithms.items()
}


def get_redis_number(key: str) -> int:
    data = redis_client.get(key)
    if data is None:
        return 0
    return int(data.decode())  # type: ignore


def get_pending_queue() -> list:
    data = redis_client.get(pending_queue_key)
    if data is None:
        return []
    return json.loads(data)


async def post_callback(
    url: str,
    uid: str,
    success: bool,
    data: Dict[str, Any],
) -> None:
    if url:
        try:
            data = dict(uid=uid, success=success, data=data)
            async with http_client.session.post(url, json=data, timeout=1) as res:
                await res.json()
        except Exception as err:
            print(
                f"\n\n!!! post to callback_url ({url}) failed "
                f"({get_err_msg(err)}) !!!\n\n"
            )


# kafka & redis
async def consume() -> None:
    OPT["verbose"] = False

    topic = kafka_topic()
    expire_seconds = 10 * 365 * 24 * 3600

    redis_client.expire(pending_queue_key, expire_seconds)
    # initialize
    http_client.start()
    for k, v in loaded_algorithms.items():
        v.initialize()
    kafka_consumer = KafkaConsumer(
        topic,
        group_id=kafka_group_id(),
        bootstrap_servers=kafka_server(),
        max_poll_records=kafka_max_poll_records(),
        max_poll_interval_ms=kafka_max_poll_interval_ms(),
    )
    # main loop
    try:
        for message in kafka_consumer:
            data = json.loads(message.value)
            uid = data["uid"]
            task = data["task"]
            params = data["params"]
            callback_url = params.get("callback_url", "")
            existing = redis_client.get(uid)
            if existing is not None:
                existing = json.loads(existing)
                print(">>> existing", existing)
                if existing["status"] in (
                    Status.FINISHED,
                    Status.EXCEPTION,
                    Status.INTERRUPTED,
                ):
                    continue
            print(">>> working", uid)
            data = {} if existing is None else (existing.get("data", {}) or {})
            start_time = time.time()
            data["start_time"] = start_time
            create_time = data.get("create_time", start_time)
            redis_client.set(uid, json.dumps(dict(status=Status.WORKING, data=data)))
            procedure = "start"
            try:
                algorithm = loaded_algorithms[task]
                model = algorithm.model_class(**params)  # type: ignore
                procedure = "start -> run_algorithm"
                res: Response = await run_algorithm(algorithm, model)
                latencies = algorithm.last_latencies
                t1 = time.time()
                procedure = "run_algorithm -> upload_temp_image"
                urls = upload_temp_image(cos_client, res.body)
                t2 = time.time()
                procedure = "upload_temp_image -> audit_image"
                if task != "img2img.sr":
                    try:
                        audit = audit_image(audit_redis_client, urls.path)
                    except:
                        audit = AuditResponse(safe=False, reason="unknown")
                else:
                    audit = AuditResponse(safe=True, reason="")
                t3 = time.time()
                procedure = "audit_image -> redis"
                result = dict(
                    cdn=urls.cdn if audit.safe else "",
                    cos=urls.cos if audit.safe else "",
                    safe=audit.safe,
                    reason=audit.reason,
                )
                result.update(data)
                end_time = time.time()
                result["end_time"] = end_time
                result["duration"] = end_time - create_time
                result["elapsed_times"] = dict(
                    pending=start_time - create_time,
                    run_algorithm=t1 - start_time,
                    algorithm_latencies=latencies,
                    upload=t2 - t1,
                    audit=t3 - t2,
                )
                result["request"] = dict(task=task, model=model.dict())
                redis_client.set(
                    uid,
                    json.dumps(dict(status=Status.FINISHED, data=result)),
                )
                t4 = time.time()
                procedure = "redis -> callback"
                await post_callback(callback_url, uid, True, result)
                t5 = time.time()
                procedure = "callback -> update_elapsed_times"
                result["elapsed_times"].update(
                    dict(
                        redis=t4 - t3,
                        callback=t5 - t4,
                    )
                )
                redis_client.set(
                    uid,
                    json.dumps(dict(status=Status.FINISHED, data=result)),
                )
                procedure = "done"
                # maintain queue
                queue = get_pending_queue()
                if uid in queue:
                    queue.remove(uid)
                    redis_client.set(pending_queue_key, json.dumps(queue))
            except Exception as err:
                end_time = time.time()
                reason = f"{task} -> {json.dumps(params, ensure_ascii=False)} -> {procedure} : {get_err_msg(err)}"
                data["reason"] = reason
                data["end_time"] = end_time
                data["duration"] = end_time - create_time
                redis_client.set(
                    uid,
                    json.dumps(dict(status=Status.EXCEPTION, data=data)),
                )
                await post_callback(callback_url, uid, False, data)
    finally:
        # clean up
        await http_client.stop()
        print(">>> end")


asyncio.run(consume())



import io
import os
import json
import time
import uuid
import logging
import requests

import numpy as np

from io import BytesIO
from PIL import Image
from enum import Enum
from typing import Union
from typing import BinaryIO
from typing import Optional
from aiohttp import ClientSession
from pydantic import Field
from pydantic import BaseModel
from cfclient.utils import download_image_with_retry as download

from .parameters import use_cos
from .parameters import bypass_audit

try:
    from redis import Redis
    from qcloud_cos import CosConfig
    from qcloud_cos import CosS3Client
except:
    Redis = None
    CosConfig = None
    CosS3Client = None


SCHEME = "http"
REGION = "ap-shanghai"
BUCKET = "ailab-1310750649"
CDN_HOST = "https://ailabcdn.nolibox.com"
COS_HOST = "https://ailab-1310750649.cos.ap-shanghai.myqcloud.com"
TEXT_BIZ_TYPE = "721ba17a279d5074cccd53587610dec4"
IMAGE_BIZ_TYPE = "f64516c2fdf256f9a5d0c45cd4e756a0"
SECRET_ID = os.getenv("SECRETID")
SECRET_KEY = os.getenv("SECRETKEY")

TEMP_TEXT_FOLDER = "tmp_txt"
TEMP_IMAGE_FOLDER = "tmp"
RETRY = 3
PART_SIZE = 20
MAX_THREAD = 1
UPLOAD_RETRY = 1
UPLOAD_TIMEOUT = 30

logger = logging.getLogger(__name__)


class UploadTextResponse(BaseModel):
    path: str = Field(..., description="The path on the cloud.")
    cdn: str = Field(..., description="The `cdn` url of the input text.")
    cos: str = Field(
        ...,
        description="The `cos` url of the input text, which should be used internally.",
    )


class AuditResponse(BaseModel):
    safe: bool = Field(..., description="Whether the input content is safe.")
    reason: str = Field(..., description="If not safe, what's the reason?")


class UploadImageResponse(BaseModel):
    path: str = Field(..., description="The path on the cloud.")
    cdn: str = Field(..., description="The `cdn` url of the input image.")
    cos: str = Field(
        ...,
        description="The `cos` url of the input image, which should be used internally.",
    )


def upload_text(
    client: CosS3Client,
    text: str,
    *,
    folder: str,
    part_size: int = 10,
    max_thread: int = 10,
) -> UploadTextResponse:
    path = f"{folder}/{uuid.uuid4().hex}.txt"
    text_io = io.StringIO(text)
    client.upload_file_from_buffer(
        BUCKET,
        path,
        text_io,
        PartSize=part_size,
        MAXThread=max_thread,
    )
    return UploadTextResponse(
        path=path,
        cdn=f"{CDN_HOST}/{path}",
        cos=f"{COS_HOST}/{path}",
    )


def upload_temp_text(
    client: CosS3Client,
    text: str,
    *,
    part_size: int = 10,
    max_thread: int = 10,
) -> UploadTextResponse:
    return upload_text(
        client,
        text,
        folder=TEMP_TEXT_FOLDER,
        part_size=part_size,
        max_thread=max_thread,
    )


def parse_audit_text(res: dict) -> Optional[AuditResponse]:
    detail = res["JobsDetail"]
    if detail["State"] != "Success":
        return
    label = detail["Label"]
    return AuditResponse(safe=label == "Normal", reason=label)


def audit_text(client: CosS3Client, text: str) -> AuditResponse:
    res = client.ci_auditing_text_submit(
        BUCKET,
        "",
        Content=text.encode("utf-8"),
        BizType=TEXT_BIZ_TYPE,
    )
    job_id = res["JobsDetail"]["JobId"]
    parsed = parse_audit_text(res)
    patience = 20
    interval = 100
    for i in range(patience):
        if parsed is not None:
            break
        time.sleep(interval)
        res = client.ci_auditing_text_query(BUCKET, job_id)
        parsed = parse_audit_text(res)
    if parsed is None:
        return AuditResponse(safe=False, reason=f"Timeout ({patience * interval})")
    return parsed


def upload_image(
    client: CosS3Client,
    inp: Union[bytes, np.ndarray, BinaryIO],
    *,
    folder: str,
    retry: int = UPLOAD_RETRY,
    timeout: int = UPLOAD_TIMEOUT,
    # part_size: int = PART_SIZE,
    # max_thread: int = MAX_THREAD,
) -> UploadImageResponse:
    path = f"{folder}/{uuid.uuid4().hex}.png"
    if isinstance(inp, bytes):
        img_bytes = io.BytesIO(inp)
    elif isinstance(inp, np.ndarray):
        img_bytes = io.BytesIO()
        Image.fromarray(inp).save(img_bytes, "PNG")
        img_bytes.seek(0)
    else:
        img_bytes = inp

    cdn_url = f"{CDN_HOST}/{path}"
    cos_url = f"{COS_HOST}/{path}"

    original_retry = client._retry
    original_timeout = client._conf._timeout
    client._retry = retry
    client._conf._timeout = timeout
    try:
        client.put_object(
            Key=path,
            Body=img_bytes,
            Bucket=BUCKET,
            StorageClass="STANDARD",
        )
    except Exception:
        try:
            raw_data = requests.get(cos_url).content
            Image.open(BytesIO(raw_data)).verify()
            logger.info("\n\ntried to get url after exception and succeeded!!!\n\n")
        except Exception:
            logger.exception("\n\ntried to get url after exception but failed!!!\n\n")
            raise
    finally:
        client._retry = original_retry
        client._conf._timeout = original_timeout
    return UploadImageResponse(
        path=path,
        cdn=cdn_url,
        cos=cos_url,
    )


def upload_temp_image(
    client: CosS3Client,
    inp: Union[bytes, np.ndarray, BinaryIO],
    # *,
    # part_size: int = PART_SIZE,
    # max_thread: int = MAX_THREAD,
) -> UploadImageResponse:
    return upload_image(
        client,
        inp,
        folder=TEMP_IMAGE_FOLDER,
        # part_size=part_size,
        # max_thread=max_thread,
    )


class ForbidEnum(int, Enum):
    NONE = 0
    FROZEN = 1
    MIGRATED = 2


class AuditJobsDetailModel(BaseModel):
    Object: str
    Label: str
    Category: str
    SubLabel: str
    ForbidState: ForbidEnum


def audit_image(audit_client: Redis, path: str, timeout: int = 3) -> AuditResponse:
    if bypass_audit():
        return AuditResponse(safe=True, reason="")
    t = time.time()
    while time.time() - t <= timeout:
        res = audit_client.get(path)
        if res is None:
            time.sleep(0.1)
            continue
        data = AuditJobsDetailModel(**json.loads(res))
        if data.Label == "Normal" and data.ForbidState == ForbidEnum.NONE:
            return AuditResponse(safe=True, reason="")
        reason = data.Label
        if data.Category:
            reason = f"{reason}/{data.Category}"
        if data.SubLabel:
            reason = f"{reason}/{data.SubLabel}"
        return AuditResponse(safe=False, reason=reason)
    return AuditResponse(safe=False, reason="timeout")


async def download_image_with_retry(
    session: ClientSession,
    url: str,
    *,
    retry: int = RETRY,
    interval: int = 1,
) -> Image.Image:
    if use_cos() and url.startswith(CDN_HOST):
        url = url.replace(CDN_HOST, COS_HOST)
    return await download(session, url, retry=retry, interval=interval)


__all__ = [
    "SCHEME",
    "REGION",
    "SECRET_ID",
    "SECRET_KEY",
    "upload_text",
    "upload_temp_text",
    "audit_text",
    "upload_image",
    "upload_temp_image",
    "audit_image",
    "ForbidEnum",
    "AuditJobsDetailModel",
    "AuditResponse",
    "UploadImageResponse",
]


if __name__ == "__main__":
    import redis
    from cfcreator.parameters import audit_redis_kwargs

    config = CosConfig(Region=REGION, SecretId=SECRET_ID, SecretKey=SECRET_KEY)
    cos_client = CosS3Client(config)
    redis_client = redis.Redis(**audit_redis_kwargs())
    print(">>> start uploading")
    # res = upload_temp_image(cos_client, np.zeros([64, 64], np.uint8))
    res = upload_temp_image(cos_client, np.array(Image.open("test.jpg")))
    print(res)
    print(">>> audit image")
    print(audit_image(redis_client, res.path))
    print(">>> upload completed")



import os
import click
import uvicorn

from cfcreator import opt_env_context


@click.group()
def main() -> None:
    pass


@main.command()
@click.option(
    "-p",
    "--port",
    default=8123,
    show_default=True,
    type=int,
    help="Port of the service.",
)
@click.option(
    "--save_gpu_ram",
    is_flag=True,
    default=False,
    show_default=True,
    type=bool,
    help="This flag will put all the models to RAM instead of GPU RAM, and only put models to GPU when they are being used.",
)
@click.option(
    "--focus",
    default="all",
    show_default=True,
    type=click.Choice(["all", "sd", "sd.base", "sd.anime", "sd.inpainting"]),
    help="""
Indicates which endpoints should we focus on, helpful if we only care about certain subset of features.
\n- all            |  will load all endpoints.
\n- sd             |  will load SD endpoints.
\n- sd.base        |  will only load basic SD endpoints, which means anime / inpainting / outpainting endpoints will not be loaded.
\n- sd.anime       |  will only load anime SD endpoints, which means basic / inpainting / outpainting endpoints will not be loaded.
\n- sd.inpainting  |  will only load inpainting / outpainting SD endpoints, which means basic / anime endpoints will not be loaded.
\n-
""",
)
@click.option(
    "-r",
    "--reload",
    is_flag=True,
    default=False,
    show_default=True,
    type=bool,
    help="This flag represents the `reload` argument of `uvicorn.run`.",
)
def serve(*, port: int, save_gpu_ram: bool, focus: str, reload: bool) -> None:
    increment = {}
    if save_gpu_ram:
        increment["save_gpu_ram"] = True
    if focus != "all":
        increment["focus"] = focus
    with opt_env_context(increment):
        uvicorn.run(
            "cfcreator.apis.interface:app",
            host="0.0.0.0",
            port=port,
            reload=reload,
            reload_dirs=os.path.dirname(__file__) if reload else None,
        )



import time

from enum import Enum
from typing import Any
from fastapi import Response
from pydantic import Field
from cfclient.models import ImageModel

from .cos import download_image_with_retry
from .common import cleanup
from .common import init_sd_ms
from .common import get_sd_from
from .common import get_sd_inpainting
from .common import handle_diffusion_model
from .common import get_bytes_from_diffusion
from .common import IAlgorithm
from .common import Txt2ImgModel
from .parameters import save_gpu_ram


txt2img_sd_endpoint = "/txt2img/sd"
txt2img_sd_inpainting_endpoint = "/txt2img/sd.inpainting"
txt2img_sd_outpainting_endpoint = "/txt2img/sd.outpainting"


class Txt2ImgSDModel(Txt2ImgModel):
    w: int = Field(512, description="The desired output width.")
    h: int = Field(512, description="The desired output height.")


@IAlgorithm.auto_register()
class Txt2ImgSD(IAlgorithm):
    model_class = Txt2ImgSDModel

    endpoint = txt2img_sd_endpoint

    def initialize(self) -> None:
        self.ms = init_sd_ms()

    async def run(self, data: Txt2ImgSDModel, *args: Any) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        m = get_sd_from(self.ms, data)
        t1 = time.time()
        size = data.w, data.h
        kwargs = handle_diffusion_model(m, data)
        img_arr = m.txt2img(
            data.text,
            size=size,
            max_wh=data.max_wh,
            **kwargs,
        ).numpy()[0]
        content = get_bytes_from_diffusion(img_arr)
        t2 = time.time()
        cleanup(m)
        self.log_times(
            {
                "get_model": t1 - t0,
                "inference": t2 - t1,
                "cleanup": time.time() - t2,
            }
        )
        return Response(content=content, media_type="image/png")


class PaddingModes(str, Enum):
    CV2_NS = "cv2_ns"
    CV2_TELEA = "cv2_telea"


class Txt2ImgSDInpaintingModel(Txt2ImgModel, ImageModel):
    mask_url: str = Field(
        ...,
        description="""
The `cdn` / `cos` url of the user's mask.
> `cos` url from `qcloud` is preferred.
> If empty string is provided, then we will use an empty mask, which means we will simply perform an image-to-image transform.  
""",
    )
    use_raw_inpainting: bool = Field(
        False,
        description="""
Whether use the raw inpainting method.
> This is useful when you want to apply inpainting with custom SD models.
""",
    )
    raw_inpainting_fidelity: float = Field(
        0.2,
        ge=0.0,
        le=1.0,
        description="The fidelity of the input image when using raw inpainting.",
    )


class Txt2ImgSDOutpaintingModel(Txt2ImgModel, ImageModel):
    pass


@IAlgorithm.auto_register()
class Txt2ImgSDInpainting(IAlgorithm):
    model_class = Txt2ImgSDInpaintingModel
    sd_inpainting_key = "$inpainting$"

    endpoint = txt2img_sd_inpainting_endpoint

    def initialize(self) -> None:
        self.ms = init_sd_ms()
        self.ms[self.sd_inpainting_key] = get_sd_inpainting()

    async def run(self, data: Txt2ImgSDInpaintingModel, *args: Any) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        image = await download_image_with_retry(self.http_client.session, data.url)
        mask = await download_image_with_retry(self.http_client.session, data.mask_url)
        m = self.ms[data.version if data.use_raw_inpainting else self.sd_inpainting_key]
        t1 = time.time()
        if save_gpu_ram():
            m.to("cuda:0", use_half=True)
        t2 = time.time()
        kwargs = handle_diffusion_model(m, data)
        img_arr = m.txt2img_inpainting(
            data.text,
            image,
            mask,
            anchor=64,
            max_wh=data.max_wh,
            use_raw_inpainting=data.use_raw_inpainting,
            raw_inpainting_fidelity=data.raw_inpainting_fidelity,
            **kwargs,
        ).numpy()[0]
        content = get_bytes_from_diffusion(img_arr)
        t3 = time.time()
        cleanup(m)
        self.log_times(
            {
                "download": t1 - t0,
                "get_model": t2 - t1,
                "inference": t3 - t2,
                "cleanup": time.time() - t3,
            }
        )
        return Response(content=content, media_type="image/png")


@IAlgorithm.auto_register()
class Txt2ImgSDOutpainting(IAlgorithm):
    model_class = Txt2ImgSDOutpaintingModel

    endpoint = txt2img_sd_outpainting_endpoint

    def initialize(self) -> None:
        self.m = get_sd_inpainting()

    async def run(self, data: Txt2ImgSDOutpaintingModel, *args: Any) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        image = await download_image_with_retry(self.http_client.session, data.url)
        t1 = time.time()
        if save_gpu_ram():
            self.m.to("cuda:0", use_half=True)
        t2 = time.time()
        kwargs = handle_diffusion_model(self.m, data)
        img_arr = self.m.outpainting(
            data.text,
            image,
            anchor=64,
            max_wh=data.max_wh,
            **kwargs,
        ).numpy()[0]
        content = get_bytes_from_diffusion(img_arr)
        t3 = time.time()
        cleanup(self.m)
        self.log_times(
            {
                "download": t1 - t0,
                "get_model": t2 - t1,
                "inference": t3 - t2,
                "cleanup": time.time() - t3,
            }
        )
        return Response(content=content, media_type="image/png")


__all__ = [
    "txt2img_sd_endpoint",
    "txt2img_sd_inpainting_endpoint",
    "txt2img_sd_outpainting_endpoint",
    "Txt2ImgSDModel",
    "Txt2ImgSDInpaintingModel",
    "Txt2ImgSDOutpaintingModel",
    "Txt2ImgSD",
    "Txt2ImgSDInpainting",
    "Txt2ImgSDOutpainting",
]



import time

import numpy as np

from PIL import Image
from typing import Any
from typing import Dict
from typing import List
from fastapi import Response
from pydantic import Field
from scipy.interpolate import NearestNDInterpolator
from cfcv.misc.toolkit import to_rgb

from .cos import download_image_with_retry
from .common import cleanup
from .common import get_esr
from .common import init_sd_ms
from .common import get_sd_from
from .common import get_semantic
from .common import get_esr_anime
from .common import get_inpainting
from .common import handle_diffusion_model
from .common import get_bytes_from_diffusion
from .common import get_bytes_from_translator
from .common import IAlgorithm
from .common import Img2ImgModel
from .common import CallbackModel
from .common import Img2ImgDiffusionModel
from .parameters import verbose
from .parameters import save_gpu_ram


img2img_sd_endpoint = "/img2img/sd"
img2img_sr_endpoint = "/img2img/sr"
img2img_inpainting_endpoint = "/img2img/inpainting"
img2img_semantic2img_endpoint = "/img2img/semantic2img"


class Img2ImgSDModel(Img2ImgDiffusionModel):
    text: str = Field(..., description="The text that we want to handle.")
    fidelity: float = Field(
        0.2,
        ge=0.0,
        le=1.0,
        description="The fidelity of the input image.",
    )
    keep_alpha: bool = Field(
        True,
        description="""
Whether the returned image should keep the alpha-channel of the input image or not.
> If the input image is a sketch image, then `keep_alpha` needs to be False in most of the time.  
""",
    )


@IAlgorithm.auto_register()
class Img2ImgSD(IAlgorithm):
    model_class = Img2ImgSDModel

    endpoint = img2img_sd_endpoint

    def initialize(self) -> None:
        self.ms = init_sd_ms()

    async def run(self, data: Img2ImgSDModel, *args: Any) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        image = await download_image_with_retry(self.http_client.session, data.url)
        t1 = time.time()
        if not data.keep_alpha:
            image = to_rgb(image)
        m = get_sd_from(self.ms, data)
        t2 = time.time()
        kwargs = handle_diffusion_model(m, data)
        img_arr = m.img2img(
            image,
            cond=[data.text],
            max_wh=data.max_wh,
            fidelity=data.fidelity,
            anchor=64,
            **kwargs,
        ).numpy()[0]
        content = get_bytes_from_diffusion(img_arr)
        t3 = time.time()
        cleanup(m)
        self.log_times(
            {
                "download": t1 - t0,
                "get_model": t2 - t1,
                "inference": t3 - t2,
                "cleanup": time.time() - t3,
            }
        )
        return Response(content=content, media_type="image/png")


class Img2ImgSRModel(Img2ImgModel, CallbackModel):
    is_anime: bool = Field(
        False,
        description="Whether the input image is an anime image or not.",
    )


@IAlgorithm.auto_register()
class Img2ImgSR(IAlgorithm):
    model_class = Img2ImgSRModel

    endpoint = img2img_sr_endpoint

    def initialize(self) -> None:
        self.esr = get_esr()
        self.esr_anime = get_esr_anime()

    async def run(self, data: Img2ImgSRModel, *args: Any) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        image = await download_image_with_retry(self.http_client.session, data.url)
        t1 = time.time()
        m = self.esr_anime if data.is_anime else self.esr
        if save_gpu_ram():
            m.to("cuda:0", use_half=True)
        t2 = time.time()
        img_arr = m.sr(image, max_wh=data.max_wh).numpy()[0]
        content = get_bytes_from_translator(img_arr)
        t3 = time.time()
        cleanup(m)
        self.log_times(
            {
                "download": t1 - t0,
                "get_model": t2 - t1,
                "inference": t3 - t2,
                "cleanup": time.time() - t3,
            }
        )
        return Response(content=content, media_type="image/png")


class Img2ImgInpaintingModel(Img2ImgDiffusionModel):
    use_refine: bool = Field(False, description="Whether should we perform refining.")
    refine_fidelity: float = Field(
        0.8,
        description="""
Refine fidelity used in inpainting.
> Only take effects when `use_refine` is set to True.
""",
    )
    mask_url: str = Field(
        ...,
        description="""
The `cdn` / `cos` url of the user's mask.
> `cos` url from `qcloud` is preferred.
> If empty string is provided, then we will use an empty mask, which means we will simply perform an image-to-image transform.  
""",
    )


@IAlgorithm.auto_register()
class Img2ImgInpainting(IAlgorithm):
    model_class = Img2ImgInpaintingModel

    endpoint = img2img_inpainting_endpoint

    def initialize(self) -> None:
        self.m = get_inpainting()

    async def run(self, data: Img2ImgInpaintingModel, *args: Any) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        image = await download_image_with_retry(self.http_client.session, data.url)
        mask_url = data.mask_url
        if not mask_url:
            mask = Image.new("L", image.size, color=0)
        else:
            mask = await download_image_with_retry(self.http_client.session, mask_url)
        t1 = time.time()
        if save_gpu_ram():
            self.m.to("cuda:0", use_half=True)
        t2 = time.time()
        refine_fidelity = data.refine_fidelity if data.use_refine else None
        img_arr = self.m.inpainting(
            image,
            mask,
            max_wh=data.max_wh,
            refine_fidelity=refine_fidelity,
            verbose=verbose(),
        ).numpy()[0]
        content = get_bytes_from_diffusion(img_arr)
        t3 = time.time()
        cleanup(self.m)
        self.log_times(
            {
                "download": t1 - t0,
                "get_model": t2 - t1,
                "inference": t3 - t2,
                "cleanup": time.time() - t3,
            }
        )
        return Response(content=content, media_type="image/png")


class Img2ImgSemantic2ImgModel(Img2ImgDiffusionModel):
    color2label: Dict[str, int] = Field(
        ...,
        description="""
Mapping of color -> (semantic) label.
> The color should be of `rgb(r,g,b)` format.
""",
    )
    keep_alpha: bool = Field(
        False,
        description="Whether the returned image should keep the alpha-channel of the input image or not.",
    )


def color2rgb(color: str) -> List[int]:
    if not color.startswith("rgb(") or not color.endswith(")"):
        raise ValueError("`color` should be of `rgb(r,g,b)` format")
    return [int(n.strip()) for n in color[4:-1].split(",")]


@IAlgorithm.auto_register()
class Img2ImgSemantic2Img(IAlgorithm):
    model_class = Img2ImgSemantic2ImgModel

    endpoint = img2img_semantic2img_endpoint

    def initialize(self) -> None:
        self.m = get_semantic()

    async def run(self, data: Img2ImgSemantic2ImgModel, *args: Any) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        raw_semantic = await download_image_with_retry(
            self.http_client.session,
            data.url,
        )
        t1 = time.time()
        w, h = raw_semantic.size
        raw_arr = np.array(raw_semantic)
        alpha = None
        valid_mask = None
        valid_mask_ravel = None
        # handle alpha
        if raw_arr.shape[-1] == 4:
            alpha = raw_arr[..., -1]
            raw_arr = raw_arr[..., :3]
            valid_mask = alpha > 0
            valid_mask_ravel = valid_mask.ravel()
        # get nearest color
        raw_arr_flat = raw_arr.reshape([h * w, 1, 3])
        if valid_mask_ravel is not None:
            raw_arr_flat = raw_arr_flat[valid_mask_ravel]
        raw_arr_flat = raw_arr_flat.astype(np.int16)
        colors = sorted(data.color2label)
        rgbs = np.array(list(map(color2rgb, colors)), np.int16).reshape([1, -1, 3])
        diff = np.abs(raw_arr_flat - rgbs).mean(2)
        indices = np.argmin(diff, axis=1)
        # diffusion has no `unlabeled` label, so it should be COCO.label - 1
        labels = np.array([data.color2label[color] - 1 for color in colors], np.uint8)
        arr_labels = labels[indices]
        if valid_mask_ravel is None:
            semantic_arr = arr_labels
        else:
            semantic_arr = np.zeros([h, w], np.uint8).ravel()
            semantic_arr[valid_mask_ravel] = arr_labels
        # nearest interpolation
        t2 = time.time()
        if valid_mask is not None and valid_mask_ravel is not None:
            to_coordinates = lambda mask: np.array(np.nonzero(mask)).T
            valid_coordinates = to_coordinates(valid_mask)
            interpolator = NearestNDInterpolator(valid_coordinates, arr_labels)
            invalid_mask = ~valid_mask
            invalid_coordinates = to_coordinates(invalid_mask)
            semantic_arr[invalid_mask.ravel()] = interpolator(invalid_coordinates)
        # gather
        semantic_arr = semantic_arr.reshape([h, w])
        semantic = Image.fromarray(semantic_arr)
        t3 = time.time()
        if save_gpu_ram():
            self.m.to("cuda:0", use_half=True)
        t4 = time.time()
        if not data.keep_alpha:
            alpha = None
        elif alpha is not None:
            alpha = alpha[None, None].astype(np.float32) / 255.0
        img_arr = self.m.semantic2img(
            semantic,
            alpha=alpha,
            max_wh=data.max_wh,
            verbose=verbose(),
        ).numpy()[0]
        content = get_bytes_from_diffusion(img_arr)
        t5 = time.time()
        cleanup(self.m)
        self.log_times(
            {
                "download": t1 - t0,
                "preprocess": t2 - t1,
                "interpolation": t3 - t2,
                "get_model": t4 - t3,
                "inference": t5 - t4,
                "cleanup": time.time() - t5,
            }
        )
        return Response(content=content, media_type="image/png")


__all__ = [
    "img2img_sd_endpoint",
    "img2img_sr_endpoint",
    "img2img_inpainting_endpoint",
    "img2img_semantic2img_endpoint",
    "Img2ImgSDModel",
    "Img2ImgSRModel",
    "Img2ImgInpaintingModel",
    "Img2ImgSemantic2ImgModel",
    "Img2ImgSD",
    "Img2ImgSR",
    "Img2ImgInpainting",
    "Img2ImgSemantic2Img",
]



import torch

import numpy as np

from abc import ABCMeta
from enum import Enum
from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import Union
from typing import Callable
from typing import Optional
from pydantic import Field
from pydantic import BaseModel
from functools import partial
from cfclient.models import TextModel
from cfclient.models import ImageModel
from cfclient.models import AlgorithmBase
from cfcv.misc.toolkit import np_to_bytes
from cflearn.api.cv import DiffusionAPI
from cflearn.api.cv import TranslatorAPI

from .parameters import OPT
from .parameters import verbose
from .parameters import save_gpu_ram


apis = {}
init_fns = {}
api_type = Union[DiffusionAPI, TranslatorAPI]


class SDVersions(str, Enum):
    v1_5_BC = ""
    v1_5 = "v1.5"
    ANIME = "anime"
    ANIME_ANYTHING = "anime_anything"


def _get(key: str, init_fn: Callable) -> api_type:
    m = apis.get(key)
    if m is not None:
        return m
    print("> init", key)
    if save_gpu_ram():
        m = init_fn("cpu")
    else:
        m = init_fn("cuda:0", use_half=True)
    apis[key] = m
    init_fns[key] = init_fn
    return m


def get_sd_version(version: SDVersions) -> DiffusionAPI:
    return _get(f"sd_{version}", partial(DiffusionAPI.from_sd_version, version))


def get_sd_anime() -> DiffusionAPI:
    return _get("sd_anime", DiffusionAPI.from_sd_anime)


def get_sd_inpainting() -> DiffusionAPI:
    return _get("sd_inpainting", DiffusionAPI.from_sd_inpainting)


def get_esr() -> TranslatorAPI:
    return _get("esr", TranslatorAPI.from_esr)


def get_esr_anime() -> TranslatorAPI:
    return _get("esr_anime", TranslatorAPI.from_esr_anime)


def get_inpainting() -> DiffusionAPI:
    return _get("inpainting", DiffusionAPI.from_inpainting)


def get_semantic() -> DiffusionAPI:
    return _get("semantic", DiffusionAPI.from_semantic)


def get_bytes_from_translator(img_arr: np.ndarray) -> bytes:
    img_arr = img_arr.transpose([1, 2, 0])
    return np_to_bytes(img_arr)


def get_bytes_from_diffusion(img_arr: np.ndarray) -> bytes:
    img_arr = 0.5 * (img_arr + 1.0)
    img_arr = img_arr.transpose([1, 2, 0])
    return np_to_bytes(img_arr)


# API models


class CallbackModel(BaseModel):
    callback_url: str = Field("", description="callback url to post to")


class MaxWHModel(BaseModel):
    max_wh: int = Field(1024, description="The maximum resolution.")


class VariationModel(BaseModel):
    seed: int = Field(..., description="Seed of the variation.")
    strength: float = Field(
        ...,
        ge=0.0,
        le=1.0,
        description="Strength of the variation.",
    )


class SDSamplers(str, Enum):
    DDIM = "ddim"
    PLMS = "plms"
    KLMS = "klms"
    SOLVER = "solver"
    K_EULER = "k_euler"
    K_EULER_A = "k_euler_a"
    K_HEUN = "k_heun"


class DiffusionModel(CallbackModel):
    use_circular: bool = Field(
        False,
        description="Whether should we use circular pattern (e.g. generate textures).",
    )
    seed: int = Field(
        -1,
        ge=-1,
        lt=2**32,
        description="""
Seed of the generation.
> If `-1`, then random seed will be used.
""",
    )
    variation_seed: int = Field(
        0,
        ge=0,
        lt=2**32,
        description="""
Seed of the variation generation.
> Only take effects when `variation_strength` is larger than 0.
""",
    )
    variation_strength: float = Field(
        0.0,
        ge=0.0,
        le=1.0,
        description="Strength of the variation generation.",
    )
    variations: List[VariationModel] = Field([], description="Variation ingredients")
    num_steps: int = Field(20, description="Number of sampling steps", ge=5, le=100)
    guidance_scale: float = Field(
        7.5,
        description="Guidance scale for classifier-free guidance.",
    )
    negative_prompt: str = Field(
        "",
        description="Negative prompt for classifier-free guidance.",
    )
    is_anime: bool = Field(
        False,
        description="Whether should we generate anime images or not.",
    )
    version: SDVersions = Field(
        SDVersions.v1_5,
        description="Version of the diffusion model",
    )
    sampler: SDSamplers = Field(
        SDSamplers.K_EULER,
        description="Sampler of the diffusion model",
    )
    clip_skip: int = Field(
        -1,
        ge=-1,
        le=8,
        description="""
Number of CLIP layers that we want to skip.
> If it is set to `-1`, then `clip_skip` = 1 if `is_anime` else 0.
""",
    )
    custom_embeddings: Dict[str, List[List[float]]] = Field(
        {},
        description="Custom embeddings, often used in textual inversion.",
    )


class Txt2ImgModel(TextModel, MaxWHModel, DiffusionModel):
    pass


class Img2ImgModel(ImageModel, MaxWHModel):
    pass


class Img2ImgDiffusionModel(Img2ImgModel, DiffusionModel):
    pass


def handle_diffusion_model(m: DiffusionAPI, data: DiffusionModel) -> Dict[str, Any]:
    seed = None if data.seed == -1 else data.seed
    variation_seed = None
    variation_strength = None
    if data.variation_strength > 0:
        variation_seed = data.variation_seed
        variation_strength = data.variation_strength
    if data.variations is None:
        variations = None
    else:
        variations = [(v.seed, v.strength) for v in data.variations]
    m.switch_circular(data.use_circular)
    unconditional_cond = [data.negative_prompt] if data.negative_prompt else None
    clip_skip = data.clip_skip
    if clip_skip == -1:
        clip_skip = 1 if data.is_anime else 0
    return dict(
        seed=seed,
        variation_seed=variation_seed,
        variation_strength=variation_strength,
        variations=variations,
        num_steps=data.num_steps,
        unconditional_guidance_scale=data.guidance_scale,
        unconditional_cond=unconditional_cond,
        sampler=data.sampler,
        verbose=verbose(),
        clip_skip=clip_skip,
        custom_embeddings=data.custom_embeddings or None,
    )


class GetPromptModel(BaseModel):
    text: str
    need_translate: bool = Field(
        True,
        description="Whether we need to translate the input text.",
    )


class GetPromptResponse(BaseModel):
    text: str
    success: bool
    reason: str


def endpoint2algorithm(endpoint: str) -> str:
    return endpoint[1:].replace("/", ".")


class IAlgorithm(AlgorithmBase, metaclass=ABCMeta):
    model_class: Type[BaseModel]
    last_latencies: Dict[str, float] = {}

    @classmethod
    def auto_register(cls) -> Callable[[AlgorithmBase], AlgorithmBase]:
        def _register(cls_: AlgorithmBase) -> AlgorithmBase:
            return cls.register(endpoint2algorithm(cls_.endpoint))(cls_)

        return _register

    def log_times(self, latencies: Dict[str, float]) -> None:
        super().log_times(latencies)
        self.last_latencies = latencies


# kafka


class Status(str, Enum):
    PENDING = "pending"
    WORKING = "working"
    FINISHED = "finished"
    EXCEPTION = "exception"
    INTERRUPTED = "interrupted"
    NOT_FOUND = "not_found"


# shortcuts


class SDParameters(BaseModel):
    is_anime: bool
    version: SDVersions


def init_sd_ms() -> Dict[str, DiffusionAPI]:
    focus = OPT.get("focus", "all")
    ms = {}
    if focus in ("all", "sd", "sd.base"):
        ms.update(
            {
                "": get_sd_version("v1.5"),
                "v1.5": get_sd_version("v1.5"),
            }
        )
    if focus in ("all", "sd", "sd.anime"):
        ms["anime"] = get_sd_anime()
        ms["anime_anything"] = get_sd_version("anime_anything_v3")
    return ms


def get_sd_from(ms: Dict[str, DiffusionAPI], data: SDParameters) -> DiffusionAPI:
    if not data.is_anime:
        m = ms[data.version]
    else:
        if data.version == SDVersions.ANIME_ANYTHING:
            m = ms["anime_anything"]
        else:
            m = ms["anime"]
    if save_gpu_ram():
        m.to("cuda:0", use_half=True)
    return m


def cleanup(m: DiffusionAPI) -> None:
    if save_gpu_ram():
        m.to("cpu")
        torch.cuda.empty_cache()


def get_api(key: str) -> Optional[api_type]:
    return apis.get(key)


def get_init_fn(key: str) -> Optional[Callable]:
    return init_fns.get(key)


def available_apis() -> List[str]:
    return sorted(apis)


__all__ = [
    "endpoint2algorithm",
    "DiffusionModel",
    "Txt2ImgModel",
    "Img2ImgModel",
    "Img2ImgDiffusionModel",
    "GetPromptModel",
    "GetPromptResponse",
    "Status",
    "IAlgorithm",
    "get_api",
    "get_init_fn",
    "available_apis",
]



import os
import json

from typing import Any
from typing import Dict
from fastapi import Response
from cftool.misc import shallow_copy_dict


OPT = dict(
    verbose=True,
    save_gpu_ram=False,
    use_cos=True,
    request_domain="localhost",
    redis_kwargs=dict(host="localhost", port=6379, db=0),
    audit_redis_kwargs=dict(host="172.17.16.7", port=6379, db=1),
    bypass_audit=False,
    kafka_server="172.17.16.8:9092",
    kafka_topic="creator",
    kafka_group_id="creator-consumer-1",
    kafka_max_poll_records=1,
    kafka_max_poll_interval_ms=5 * 60 * 1000,
    pending_queue_key="KAFKA_PENDING_QUEUE",
)
OPT_ENV_KEY = "CFCREATOR_ENV"


class opt_context:
    def __init__(self, increment: Dict[str, Any]) -> None:
        self._increment = increment
        self._backup = shallow_copy_dict(OPT)

    def __enter__(self) -> None:
        OPT.update(self._increment)

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        OPT.update(self._backup)


class opt_env_context:
    def __init__(self, increment: Dict[str, Any]) -> None:
        self._increment = increment
        self._backup = os.environ.get(OPT_ENV_KEY)

    def __enter__(self) -> None:
        os.environ[OPT_ENV_KEY] = json.dumps(self._increment)

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        if self._backup is None:
            del os.environ[OPT_ENV_KEY]
        else:
            os.environ[OPT_ENV_KEY] = self._backup


def verbose() -> bool:
    return OPT["verbose"]


def save_gpu_ram() -> bool:
    return OPT["save_gpu_ram"]


def use_cos() -> bool:
    return OPT["use_cos"]


def inject_headers(response: Response) -> None:
    response.headers["X-Request-Domain"] = OPT["request_domain"]


def redis_kwargs() -> Dict[str, Any]:
    return shallow_copy_dict(OPT["redis_kwargs"])


def audit_redis_kwargs() -> Dict[str, Any]:
    return shallow_copy_dict(OPT["audit_redis_kwargs"])


def bypass_audit() -> bool:
    return OPT["bypass_audit"]


def kafka_server() -> str:
    return OPT["kafka_server"]


def kafka_topic() -> str:
    return OPT["kafka_topic"]


def kafka_group_id() -> str:
    return OPT["kafka_group_id"]


def kafka_max_poll_records() -> int:
    return OPT["kafka_max_poll_records"]


def kafka_max_poll_interval_ms() -> int:
    return OPT["kafka_max_poll_interval_ms"]


def get_pending_queue_key() -> str:
    return OPT["pending_queue_key"]


__all__ = [
    "OPT",
    "OPT_ENV_KEY",
    "opt_context",
    "opt_env_context",
    "use_cos",
    "verbose",
    "save_gpu_ram",
    "inject_headers",
    "redis_kwargs",
    "audit_redis_kwargs",
    "kafka_server",
    "kafka_topic",
    "kafka_group_id",
    "kafka_max_poll_records",
    "kafka_max_poll_interval_ms",
    "get_pending_queue_key",
]
